class RioDeprecationWarning(DeprecationWarning):
    """
    The user used functionality that has been deprecated.
    """


class RioPotentialMistakeWarning(Warning):
    """
    The user did something sketchy that's probably a bad idea.
    """
